/*
 * CONVEKS - Copyright (C) 2018 E. Kuci, C. Geuzaine and P. Duysinx, University of Liege
 *
 * See the LICENSE.txt file for license information. Please report all
 * bugs and problems to the public mailing list <conveks@onelab.info>.
 */

#ifndef _CONVEKSC_H_
#define _CONVEKSC_H_

/*
 * This file defines the CONVEKS C API (v1.1).
 *
 * Do not edit it directly: it is automatically generated by `api/gen.py'.
 *
 * By design, the CONVEKS C API is purely functional, and only uses elementary
 * types. See `demos/api' for examples.
 */

#include <stddef.h>

#define CONVEKS_API_VERSION "1.1"

#if defined(CONVEKS_DLL)
#if defined(CONVEKS_DLL_EXPORT)
#define CONVEKS_API __declspec(dllexport)
#else
#define CONVEKS_API __declspec(dllimport)
#endif
#else
#define CONVEKS_API
#endif

CONVEKS_API void conveksFree(void *p);
CONVEKS_API void *conveksMalloc(size_t n);

/* Initialize Conveks. */
CONVEKS_API void conveksInitialize(const char * type,
                                   int * ierr);

/* Solve the optimization problem. */
CONVEKS_API void conveksRun(int * ierr);

/* Add the objective function. */
CONVEKS_API void conveksAddObjective(const void * func,
                                     int * ierr);

/* Solve the optimization problem. */
CONVEKS_API void conveksAddConstraint(int * ierr);

/* Solve the optimization problem. */
CONVEKS_API void conveksAddGroupOfConstraints(int * ierr);

/* Solve the optimization problem. */
CONVEKS_API void conveksAddVariable(int * ierr);

/* Solve the optimization problem. */
CONVEKS_API void conveksAddGroupOfVariables(int * ierr);

/* Compute the objective function. */
CONVEKS_API void conveksComputeObjective(int * ierr);

/* Finalize Conveks. */
CONVEKS_API void conveksFinalize(int * ierr);

/* Set a numerical option to `value'. `name'. Available categories and options
 * are listed in the Conveks reference manual. */
CONVEKS_API void conveksOptionSetNumber(const char * name,
                                        const double value,
                                        int * ierr);

/* Get the `value' of a numerical option. */
CONVEKS_API void conveksOptionGetNumber(const char * name,
                                        double * value,
                                        int * ierr);

/* Get all the options. */
CONVEKS_API void conveksOptionGetList(char *** names, size_t * names_n,
                                      int * ierr);

/* Print the numerical options. */
CONVEKS_API void conveksOptionShow(int * ierr);

/* Initialize MMA. */
CONVEKS_API void conveksMmaInitialize(double * currentPoint, size_t currentPoint_n,
                                      double * lowerBound, size_t lowerBound_n,
                                      double * upperBound, size_t upperBound_n,
                                      const int restart,
                                      int * ierr);

/* Get current point. */
CONVEKS_API void conveksMmaGetCurrentPoint(double ** currentPoint, size_t * currentPoint_n,
                                           int * ierr);

/* Update current point (based on first order derivatives and possibly the
 * diagonal entries of the objective function hessian). */
CONVEKS_API void conveksMmaUpdateCurrentPoint(const double objective,
                                              double * gradientOfObjective, size_t gradientOfObjective_n,
                                              double * constraints, size_t constraints_n,
                                              double * gradientOfConstraints, size_t gradientOfConstraints_n,
                                              double * diagHessianOfObjective, size_t diagHessianOfObjective_n,
                                              const double objectiveAtUpdatedPoint,
                                              double * constraintsAtUpdatedPoint, size_t constraintsAtUpdatedPoint_n,
                                              int * ierr);

/* Get the norm of KKT at current point. */
CONVEKS_API void conveksMmaGetKKTNorm(double * norm,
                                      double * constraints, size_t constraints_n,
                                      double * gradientOfObjective, size_t gradientOfObjective_n,
                                      double * gradientOfConstraints, size_t gradientOfConstraints_n,
                                      int * ierr);

/* Check if the submodel is conservative and return 1 if it is the case,
 * otherwise return 0. */
CONVEKS_API void conveksMmaIsConservativeSubProblem(int * conservative,
                                                    const double objective,
                                                    double * constraints, size_t constraints_n,
                                                    int * ierr);

/* Get outer iteration. */
CONVEKS_API void conveksMmaGetOuterIteration(int * iteration,
                                             int * ierr);

/* Get total number of nonlinear iterations. */
CONVEKS_API void conveksMmaCountNonlinearIterations(int * iteration,
                                                    int * ierr);

/* Get the infinite norm between two successive designs. */
CONVEKS_API void conveksMmaGetDesignChange(double * change,
                                           int * ierr);

/* Set the outer move limits. */
CONVEKS_API void conveksMmaSetMoveLimits(double * lowerBound, size_t lowerBound_n,
                                         double * upperBound, size_t upperBound_n,
                                         const double move,
                                         int * ierr);

/* Finalize MMA. */
CONVEKS_API void conveksMmaFinalize(int * ierr);

/* Set a numerical option to `value'. `name'. Available categories and options
 * are listed in the MMA reference manual. */
CONVEKS_API void conveksMmaOptionSetNumber(const char * name,
                                           const double value,
                                           int * ierr);

/* Get the `value' of a numerical option. */
CONVEKS_API void conveksMmaOptionGetNumber(const char * name,
                                           double * value,
                                           int * ierr);

/* Print the numerical options. */
CONVEKS_API void conveksMmaOptionShow(int * ierr);

#endif
