# CONVEKS - Copyright (C) 2018 E. Kuci, C. Geuzaine and P. Duysinx, University of Liege
#
# See the LICENSE.txt file for license information. Please report all
# bugs and problems to the public mailing list <conveks@onelab.info>.

# This file defines the CONVEKS Julia API (v1.1).
#
# Do not edit it directly: it is automatically generated by `api/gen.py'.
#
# By design, the CONVEKS Julia API is purely functional, and only uses elementary
# types. See `demos/api' for examples.

"""
    module conveks

Top-level functions
"""
module conveks

const CONVEKS_API_VERSION = "1.1"
const libdir = dirname(@__FILE__)
const lib = joinpath(libdir, Sys.iswindows() ? "conveks-1.1" : "libconveks")

"""
    conveks.initialize(type = "none")

Initialize Conveks.
"""
function initialize(type = "none")
    ierr = Ref{Cint}()
    ccall((:conveksInitialize, lib), Nothing,
          (Ptr{Cchar}, Ptr{Cint}),
          type, ierr)
    ierr[] != 0 && error("conveksInitialize returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.run()

Solve the optimization problem.
"""
function run()
    ierr = Ref{Cint}()
    ccall((:conveksRun, lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksRun returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.addObjective(func)

Add the objective function.
"""
function addObjective(func)
    ierr = Ref{Cint}()
    ccall((:conveksAddObjective, lib), Nothing,
          (Ptr{Nothing}, Ptr{Cint}),
          func, ierr)
    ierr[] != 0 && error("conveksAddObjective returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.addConstraint()

Solve the optimization problem.
"""
function addConstraint()
    ierr = Ref{Cint}()
    ccall((:conveksAddConstraint, lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksAddConstraint returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.addGroupOfConstraints()

Solve the optimization problem.
"""
function addGroupOfConstraints()
    ierr = Ref{Cint}()
    ccall((:conveksAddGroupOfConstraints, lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksAddGroupOfConstraints returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.addVariable()

Solve the optimization problem.
"""
function addVariable()
    ierr = Ref{Cint}()
    ccall((:conveksAddVariable, lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksAddVariable returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.addGroupOfVariables()

Solve the optimization problem.
"""
function addGroupOfVariables()
    ierr = Ref{Cint}()
    ccall((:conveksAddGroupOfVariables, lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksAddGroupOfVariables returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.computeObjective()

Compute the objective function.
"""
function computeObjective()
    ierr = Ref{Cint}()
    ccall((:conveksComputeObjective, lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksComputeObjective returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.finalize()

Finalize Conveks.
"""
function finalize()
    ierr = Ref{Cint}()
    ccall((:conveksFinalize, lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksFinalize returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    module conveks.option

Global option handling functions
"""
module option

import ..conveks

"""
    conveks.option.setNumber(name, value)

Set a numerical option to `value`. `name`. Available categories and options are
listed in the Conveks reference manual.
"""
function setNumber(name, value)
    ierr = Ref{Cint}()
    ccall((:conveksOptionSetNumber, conveks.lib), Nothing,
          (Ptr{Cchar}, Cdouble, Ptr{Cint}),
          name, value, ierr)
    ierr[] != 0 && error("conveksOptionSetNumber returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.option.getNumber(name)

Get the `value` of a numerical option.

Return `value`.
"""
function getNumber(name)
    api_value_ = Ref{Cdouble}()
    ierr = Ref{Cint}()
    ccall((:conveksOptionGetNumber, conveks.lib), Nothing,
          (Ptr{Cchar}, Ptr{Cdouble}, Ptr{Cint}),
          name, api_value_, ierr)
    ierr[] != 0 && error("conveksOptionGetNumber returned non-zero error code: $(ierr[])")
    return api_value_[]
end

"""
    conveks.option.getList()

Get all the options.

Return `names`.
"""
function getList()
    api_names_ = Ref{Ptr{Ptr{Cchar}}}()
    api_names_n_ = Ref{Csize_t}()
    ierr = Ref{Cint}()
    ccall((:conveksOptionGetList, conveks.lib), Nothing,
          (Ptr{Ptr{Cchar}}, Ptr{Csize_t}, Ptr{Cint}),
          api_names_, api_names_n_, ierr)
    ierr[] != 0 && error("conveksOptionGetList returned non-zero error code: $(ierr[])")
    tmp_api_names_ = unsafe_wrap(Array, api_names_[], api_names_n_[], own=true)
    names = [unsafe_string(tmp_api_names_[i]) for i in 1:length(tmp_api_names_) ]
    return names
end

"""
    conveks.option.show()

Print the numerical options.
"""
function show()
    ierr = Ref{Cint}()
    ccall((:conveksOptionShow, conveks.lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksOptionShow returned non-zero error code: $(ierr[])")
    return nothing
end

end # end of module option

"""
    module conveks.mma

MMA optimizer
"""
module mma

import ..conveks

"""
    conveks.mma.initialize(currentPoint, lowerBound, upperBound, restart = false)

Initialize MMA.
"""
function initialize(currentPoint, lowerBound, upperBound, restart = false)
    ierr = Ref{Cint}()
    ccall((:conveksMmaInitialize, conveks.lib), Nothing,
          (Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Cint, Ptr{Cint}),
          currentPoint, length(currentPoint), lowerBound, length(lowerBound), upperBound, length(upperBound), restart, ierr)
    ierr[] != 0 && error("conveksMmaInitialize returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.mma.getCurrentPoint()

Get current point.

Return `currentPoint`.
"""
function getCurrentPoint()
    api_currentPoint_ = Ref{Ptr{Cdouble}}()
    api_currentPoint_n_ = Ref{Csize_t}()
    ierr = Ref{Cint}()
    ccall((:conveksMmaGetCurrentPoint, conveks.lib), Nothing,
          (Ptr{Ptr{Cdouble}}, Ptr{Csize_t}, Ptr{Cint}),
          api_currentPoint_, api_currentPoint_n_, ierr)
    ierr[] != 0 && error("conveksMmaGetCurrentPoint returned non-zero error code: $(ierr[])")
    currentPoint = unsafe_wrap(Array, api_currentPoint_[], api_currentPoint_n_[], own=true)
    return currentPoint
end

"""
    conveks.mma.updateCurrentPoint(objective, gradientOfObjective, constraints = Cdouble[], gradientOfConstraints = Cdouble[], diagHessianOfObjective = Cdouble[], objectiveAtUpdatedPoint = nan, constraintsAtUpdatedPoint = Cdouble[])

Update current point (based on first order derivatives and possibly the diagonal
entries of the objective function hessian).
"""
function updateCurrentPoint(objective, gradientOfObjective, constraints = Cdouble[], gradientOfConstraints = Cdouble[], diagHessianOfObjective = Cdouble[], objectiveAtUpdatedPoint = nan, constraintsAtUpdatedPoint = Cdouble[])
    ierr = Ref{Cint}()
    ccall((:conveksMmaUpdateCurrentPoint, conveks.lib), Nothing,
          (Cdouble, Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Cdouble, Ptr{Cdouble}, Csize_t, Ptr{Cint}),
          objective, gradientOfObjective, length(gradientOfObjective), constraints, length(constraints), gradientOfConstraints, length(gradientOfConstraints), diagHessianOfObjective, length(diagHessianOfObjective), objectiveAtUpdatedPoint, constraintsAtUpdatedPoint, length(constraintsAtUpdatedPoint), ierr)
    ierr[] != 0 && error("conveksMmaUpdateCurrentPoint returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.mma.getKKTNorm(constraints, gradientOfObjective, gradientOfConstraints)

Get the norm of KKT at current point.

Return `norm`.
"""
function getKKTNorm(constraints, gradientOfObjective, gradientOfConstraints)
    api_norm_ = Ref{Cdouble}()
    ierr = Ref{Cint}()
    ccall((:conveksMmaGetKKTNorm, conveks.lib), Nothing,
          (Ptr{Cdouble}, Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Ptr{Cint}),
          api_norm_, constraints, length(constraints), gradientOfObjective, length(gradientOfObjective), gradientOfConstraints, length(gradientOfConstraints), ierr)
    ierr[] != 0 && error("conveksMmaGetKKTNorm returned non-zero error code: $(ierr[])")
    return api_norm_[]
end

"""
    conveks.mma.isConservativeSubProblem(objective, constraints = Cdouble[])

Check if the submodel is conservative and return 1 if it is the case, otherwise
return 0.

Return `conservative`.
"""
function isConservativeSubProblem(objective, constraints = Cdouble[])
    api_conservative_ = Ref{Cint}()
    ierr = Ref{Cint}()
    ccall((:conveksMmaIsConservativeSubProblem, conveks.lib), Nothing,
          (Ptr{Cint}, Cdouble, Ptr{Cdouble}, Csize_t, Ptr{Cint}),
          api_conservative_, objective, constraints, length(constraints), ierr)
    ierr[] != 0 && error("conveksMmaIsConservativeSubProblem returned non-zero error code: $(ierr[])")
    return api_conservative_[]
end

"""
    conveks.mma.getOuterIteration()

Get outer iteration.

Return `iteration`.
"""
function getOuterIteration()
    api_iteration_ = Ref{Cint}()
    ierr = Ref{Cint}()
    ccall((:conveksMmaGetOuterIteration, conveks.lib), Nothing,
          (Ptr{Cint}, Ptr{Cint}),
          api_iteration_, ierr)
    ierr[] != 0 && error("conveksMmaGetOuterIteration returned non-zero error code: $(ierr[])")
    return api_iteration_[]
end

"""
    conveks.mma.countNonlinearIterations()

Get total number of nonlinear iterations.

Return `iteration`.
"""
function countNonlinearIterations()
    api_iteration_ = Ref{Cint}()
    ierr = Ref{Cint}()
    ccall((:conveksMmaCountNonlinearIterations, conveks.lib), Nothing,
          (Ptr{Cint}, Ptr{Cint}),
          api_iteration_, ierr)
    ierr[] != 0 && error("conveksMmaCountNonlinearIterations returned non-zero error code: $(ierr[])")
    return api_iteration_[]
end

"""
    conveks.mma.getDesignChange()

Get the infinite norm between two successive designs.

Return `change`.
"""
function getDesignChange()
    api_change_ = Ref{Cdouble}()
    ierr = Ref{Cint}()
    ccall((:conveksMmaGetDesignChange, conveks.lib), Nothing,
          (Ptr{Cdouble}, Ptr{Cint}),
          api_change_, ierr)
    ierr[] != 0 && error("conveksMmaGetDesignChange returned non-zero error code: $(ierr[])")
    return api_change_[]
end

"""
    conveks.mma.setMoveLimits(lowerBound, upperBound, move)

Set the outer move limits.
"""
function setMoveLimits(lowerBound, upperBound, move)
    ierr = Ref{Cint}()
    ccall((:conveksMmaSetMoveLimits, conveks.lib), Nothing,
          (Ptr{Cdouble}, Csize_t, Ptr{Cdouble}, Csize_t, Cdouble, Ptr{Cint}),
          lowerBound, length(lowerBound), upperBound, length(upperBound), move, ierr)
    ierr[] != 0 && error("conveksMmaSetMoveLimits returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.mma.finalize()

Finalize MMA.
"""
function finalize()
    ierr = Ref{Cint}()
    ccall((:conveksMmaFinalize, conveks.lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksMmaFinalize returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    module conveks.mma.option

Global option handling functions
"""
module option

import ...conveks

"""
    conveks.mma.option.setNumber(name, value)

Set a numerical option to `value`. `name`. Available categories and options are
listed in the MMA reference manual.
"""
function setNumber(name, value)
    ierr = Ref{Cint}()
    ccall((:conveksMmaOptionSetNumber, conveks.lib), Nothing,
          (Ptr{Cchar}, Cdouble, Ptr{Cint}),
          name, value, ierr)
    ierr[] != 0 && error("conveksMmaOptionSetNumber returned non-zero error code: $(ierr[])")
    return nothing
end

"""
    conveks.mma.option.getNumber(name)

Get the `value` of a numerical option.

Return `value`.
"""
function getNumber(name)
    api_value_ = Ref{Cdouble}()
    ierr = Ref{Cint}()
    ccall((:conveksMmaOptionGetNumber, conveks.lib), Nothing,
          (Ptr{Cchar}, Ptr{Cdouble}, Ptr{Cint}),
          name, api_value_, ierr)
    ierr[] != 0 && error("conveksMmaOptionGetNumber returned non-zero error code: $(ierr[])")
    return api_value_[]
end

"""
    conveks.mma.option.show()

Print the numerical options.
"""
function show()
    ierr = Ref{Cint}()
    ccall((:conveksMmaOptionShow, conveks.lib), Nothing,
          (Ptr{Cint},),
          ierr)
    ierr[] != 0 && error("conveksMmaOptionShow returned non-zero error code: $(ierr[])")
    return nothing
end

end # end of module option

end # end of module mma

end # end of module conveks
